// ===== CUE ezPatch object =====================================================

// *** ezPatch static variables ***

var ezPatch_static_idGen = 0;

// *** Constant Values ***
var PLAY_KEYBOARD = 0;
var PLAY_GUITAR = 1;
var PLAY_FRETBOARD = 2;
var PLAY_CONTINUOUS = 3;

var OFF_VEL = 16;
var GLITCH_WINDOW = 750;


// *** ezPatch constructors ***

function ezPatch(id)
{
	if(id != null && id != "") {
		this.id = id;
	} else {
		this.id = "ezPatch" + ezPatch_static_idGen;
	}

	// Allocate space
	this.row = new Array();
	for(var ri = 0; ri < 6; ri++) {
		this.row[ri] = new Object();
		for(var fi = -1; fi < 13; fi++) {
			this.row[ri][fi] = new Object();
	  }
	}

	// Initialize this puppy!
	this.reset();

	ezPatch_static_idGen++;
}


// *** ezPatch properties with defaults ***

// Internals

ezPatch.prototype.type = "ezPatch";
ezPatch.prototype.id = "ezPatch#";

ezPatch.prototype.row = null;

// User properties

ezPatch.prototype.name = "Uninitialized ezPatch Name";
ezPatch.prototype.description = "Uninitialized ezPatch description here";
ezPatch.prototype.playStyle = PLAY_KEYBOARD;
ezPatch.prototype.openStr = true;

ezPatch.prototype.velOrig = 100;
ezPatch.prototype.velHammer = 100;
ezPatch.prototype.velPull = 100;

// *** ezPatch method declarations ***

// Base functionality

ezPatch.prototype.reset = ezPatch_fn_reset;
ezPatch.prototype.panic = ezPatch_fn_panic;

ezPatch.prototype.fretOff = ezPatch_fn_fretOff;
ezPatch.prototype.fretOn = ezPatch_fn_fretOn;
ezPatch.prototype.dropHighNote = ezPatch_fn_dropHighNote;
ezPatch.prototype.addHighNote = ezPatch_fn_addHighNote;

ezPatch.prototype.pitchWheel = ezPatch_fn_pitchWheel;
ezPatch.prototype.chaseWheel = ezPatch_fn_chaseWheel;
ezPatch.prototype.stringPluck = ezPatch_fn_stringPluck;

ezPatch.prototype.plusButtonPress = ezPatch_fn_plusButtonPress;
ezPatch.prototype.plusButtonRelease = ezPatch_fn_plusButtonRelease;
ezPatch.prototype.minusButtonPress = ezPatch_fn_minusButtonPress;
ezPatch.prototype.minusButtonRelease = ezPatch_fn_minusButtonRelease;
ezPatch.prototype.soundButtonPress = ezPatch_fn_soundButtonPress;
ezPatch.prototype.soundButtonRelease = ezPatch_fn_soundButtonRelease;
ezPatch.prototype.songButtonPress = ezPatch_fn_songButtonPress;
ezPatch.prototype.songButtonRelease = ezPatch_fn_songButtonRelease;


// *** ezPatch method implementation ***

function ezPatch_fn_reset(id)
{
	if(this.row == null) {
		// Buh?
		alert("patch not allocated?");
		return;
	}

	for(var ri = 0; ri < 6; ri++) {
		for(var fi = -1; fi < 13; fi++) {
			this.row[ri][fi].on = false;
			this.row[ri][fi].note = -1;
			this.row[ri][fi].port = -1;
			this.row[ri][fi].channel = -1;
	  }
		this.row[ri].baseNote = -1;
		this.row[ri].port = 0;
		this.row[ri].channel = 0;
		this.row[ri].highest = -1;
		this.row[ri].playing = -2;
		this.row[ri].lastPluck = 0;
		this.row[ri].velLowIn = 30;
		this.row[ri].velHighIn = 100;
	}

	this.velStringLow = 1;
	this.velStringHigh = 127;

	this.name = "Uninitialized ezPatch Name";
	this.description = "Uninitialized ezPatch description here";
	this.playStyle = PLAY_KEYBOARD;
	this.openStr = true;
}

function ezPatch_fn_panic(id)
{
	if(this.row == null) {
		// Buh?
		alert("patch not allocated?");
		return;
	}

	for(var ri = 0; ri < 6; ri++) {
		for(var fi = -1; fi < 13; fi++) {
			if(this.row[ri][fi].on) {
				// Turn note off
				if(this.row[ri].channel > 0)
				OutputMidiMsg( -1, 0x80, this.row[ri].channel,
					fi+this.row[ri].baseNote, OFF_VEL );
				this.row[ri][fi].on = false;
				guiFret(fi == -1? 0 : 2, ri, fi);
			}
	  }
	  if(this.playStyle == PLAY_CONTINUOUS && this.row[ri].playing > -2) {
			// Make sure held notes are turned off
			OutputMidiMsg( -1, 0x80, this.row[ri].channel,
				this.row[ri].playing+this.row[ri].baseNote, OFF_VEL );
			guiFret(0, ri, this.row[ri].playing);
			this.row[ri].playing == -2;
		}
	}
}


function ezPatch_fn_fretOn(row, fret)
{
	if(this.row[row].channel > 0) {
		if(this.playStyle != PLAY_KEYBOARD) {
			this.addHighNote(row, fret);
		} else {
			guiFret(1, row, fret);
			OutputMidiMsg( -1, 0x90, this.row[row].channel,
				fret+this.row[row].baseNote, this.velOrig );
			this.row[row][fret].on = true;
		}
	}
}

function ezPatch_fn_fretOff(row, fret)
{
	if(this.row[row].channel > 0) {
		if(this.playStyle != PLAY_KEYBOARD) {
			this.dropHighNote(row, fret);
		} else {
			if(this.row[row][fret].on) {
			OutputMidiMsg( -1, 0x80, this.row[row].channel,
				fret+this.row[row].baseNote, OFF_VEL );
				this.row[row][fret].on = false;
				guiFret(0, row, fret);
			}
		}
	}
}


function ezPatch_fn_addHighNote(row, fret)
{
	// Mark this note as ON
	this.row[row][fret].on = true;
	guiFret(2, row, fret)

	// Is this note the highest?
	if(fret > this.row[row].highest) {
		// Turn on this new highest
if(this.playStyle != PLAY_CONTINUOUS)
		OutputMidiMsg( -1, 0x90, this.row[row].channel,
			fret+this.row[row].baseNote,
				(this.playStyle == PLAY_FRETBOARD ? 100 :
					((this.row[row].highest != -1 || this.row[row][-1].on) ? this.velHammer : this.velOrig)));

		// Turn off previous highest
		if(this.row[row].highest != -1 || this.row[row][-1].on) {
if(this.playStyle != PLAY_CONTINUOUS)
			OutputMidiMsg( -1, 0x80, this.row[row].channel,
				this.row[row].highest+this.row[row].baseNote, OFF_VEL);
			// Dim "hidden" fret
			guiFret(2, row, this.row[row].highest);
		}

		// Remember this one as highest
		this.row[row].highest = fret;
		// Light new highest
		guiFret(1, row, fret);
	}
	// Otherwise, just remember it for later



	if(this.playStyle == PLAY_CONTINUOUS) {
		// Always display held note
		guiFret(3, row, this.row[row].playing);
	}
}


function ezPatch_fn_dropHighNote(row, fret)
{
	this.row[row][fret].on = false;
	guiFret(0, row, fret)

	// Is this our top/held/sounding note?
	if(fret == this.row[row].highest) {
		// Turn on next lowest note
		for(var il = this.row[row].highest-1; il >= -1; il--) {
			if(this.row[row][il].on) {
				// This is our new highest
if(this.playStyle != PLAY_CONTINUOUS)
				OutputMidiMsg( -1, 0x90, this.row[row].channel,
					il+this.row[row].baseNote, (this.playStyle == PLAY_FRETBOARD ? this.velOrig : this.velPull) );
				// Remember this as our highest
				this.row[row].highest = il;
				guiFret(1, row, il);
				break;
			}
		}
		// Turn it off
if(this.playStyle != PLAY_CONTINUOUS)
		OutputMidiMsg( -1, 0x80, this.row[row].channel,
			fret+this.row[row].baseNote, OFF_VEL );
		//guiFret(0, row, fret);

		if(il < 0) {
			// Apparently no other notes are still held: reset our lowest value
			this.row[row].highest = -1;
			guiVelocity(row, 0);
		}
	}



	if(this.playStyle == PLAY_CONTINUOUS) {
		// Always display held note
		guiFret(3, row, this.row[row].playing);
	}
}


function ezPatch_fn_stringPluck(row, vel)
{

	// MFF: String velocity adjustment here!

	if(vel) {
		if(vel > this.row[row].velHighIn) vel = this.row[row].velHighIn;
		vel -= this.row[row].velLowIn;
		if(vel < 0) vel = 0;

		// vel now ranges between high and low

		vel = (vel * (this.velStringHigh - this.velStringLow)) / (this.row[row].velHighIn - this.row[row].velLowIn);
		vel += this.velStringLow;

		vel = Math.round(vel);

		if(vel < 1) vel = 1;
		if(vel > 127) vel = 127;

		PrintLn("New vel = " + vel);
	}

	if(this.playStyle == PLAY_GUITAR || this.playStyle == PLAY_CONTINUOUS) {
		if(this.row[row].channel == 0) {
			// Muted string: ignore
			return;
		}

		// Check for multi-triggers
		if(Midi.GetMasterClockTime() - this.row[row].lastPluck > GLITCH_WINDOW) {
			var high = this.row[row].highest;

			if(!this.openStr && high == -1 && vel > 0) {
				// Abort: open string in "no open" mode
				return;
			}


if(this.playStyle == PLAY_GUITAR) {
			if(vel == 0) {
				// Turn off the highest note
				if(high != -1 || this.row[row][-1].on) {
					OutputMidiMsg( -1, 0x80, this.row[row].channel,
						high+this.row[row].baseNote, OFF_VEL );
				}
				guiFret(2, row, high);
				// Also make sure open notes are cancelled!
				guiFret(0, row, -1);
				this.row[row][-1].on = false;
			} else {
				// Retrigger the highest note
				if(high != -1 || this.row[row][-1].on) {
					OutputMidiMsg( -1, 0x80, this.row[row].channel,
						high+this.row[row].baseNote, OFF_VEL);
				}
				OutputMidiMsg( -1, 0x90, this.row[row].channel,
					high+this.row[row].baseNote, vel);
				this.row[row].lastPluck = Midi.GetMasterClockTime();
				guiFret(1, row, high)
			}
}


if(this.playStyle == PLAY_CONTINUOUS) {
			var play = this.row[row].playing;
			if(vel == 0) {
				// Turn off the playing note
				if(play != -1 || this.row[row][-1].on) {
					OutputMidiMsg( -1, 0x80, this.row[row].channel,
						play+this.row[row].baseNote, OFF_VEL );
				}
				guiFret(0, row, high);
				// Also make sure open notes are cancelled!
				guiFret(0, row, -1);
				this.row[row][-1].on = false;
				guiFret(0, row, play);
				this.row[row].playing = -2;
			} else {
				// Retrigger the highest note
				if(play != -1 || this.row[row][-1].on) {
					OutputMidiMsg( -1, 0x80, this.row[row].channel,
						play+this.row[row].baseNote, OFF_VEL);
				}
				if(high != play) {
					// Redraw last fret
					guiFret(0, row, play);
				}
				OutputMidiMsg( -1, 0x90, this.row[row].channel,
					high+this.row[row].baseNote, vel);
				this.row[row].lastPluck = Midi.GetMasterClockTime();
				guiFret(1, row, high)
				this.row[row].playing = high;
			}
}





			if(high == -1) {
				this.row[row][-1].on = vel ? true : false;
			}

			guiVelocity(row, vel);


if(this.playStyle == PLAY_CONTINUOUS) {
	// Always display held note
	if(high != -1 || this.row[row][-1].on)
		guiFret(3, row, this.row[row].playing);
}

		}
	} else {
		/*
		// Do what for keyboard mode?
		guiFret(1, row, fret);
		OutputMidiMsg( -1, 0x90, this.row[row].channel,
			fret+this.row[row].baseNote, 100 );
		*/
	}
}


var wheelTarget = 0x2000;
var wheelTracker = 0x2000;
var wheelStep = 500;
var wheelThread = -1;

function ezPatch_fn_pitchWheel(dat1, dat2)
{
	// 8192 == 0x2000 = centered
	// Sometimes sticks as high as 8972 on way down!
	// Sometimes sticks as low as 6656 on way up!
	var pwval = (dat2 << 7) | dat1;

	//PrintLn("PitchWheel in: " + pwval);

	if(pwval >= 6656 && pwval <= 8972) {
		if(pwval < 0x2000) {
			if(wheelTracker < pwval) {
				pwval = 0x2000;
			}
		} else if(pwval > 0x2000) {
			if(wheelTracker > pwval) {
				pwval = 0x2000;
			}
		}
	}

	wheelTarget = pwval;

	if(wheelTarget > wheelTracker) {
		wheelStep = Math.round((wheelTarget-wheelTracker)/6);
	} else {
		wheelStep = Math.round((wheelTracker-wheelTarget)/6);
	}

	if(wheelTracker != wheelTarget) {
		// Do first step immediately, not on interval
		this.chaseWheel();
	}

	if(wheelThread == -1) {
		wheelThread = setInterval("patch.chaseWheel()", 1);
		//PrintLn("PitchWheel target: " + pwval);
	}

}

function ezPatch_fn_chaseWheel()
{

	if(wheelTracker < wheelTarget) {
		wheelTracker += wheelStep;
		if(wheelTracker > wheelTarget) {
			wheelTracker = wheelTarget;
		}
	} else if(wheelTracker > wheelTarget) {
		wheelTracker -= wheelStep;
		if(wheelTracker < wheelTarget) {
			wheelTracker = wheelTarget;
		}
	} else {
		// Equal!  Shouldn't be here, but let's cancel anyway.
		if(wheelThread != -1) {
			clearInterval(wheelThread);
			wheelThread = -1;
		}
		return;
	}

	var dat1 = wheelTracker & 0x7F;
	var dat2 = wheelTracker >> 7;

	var arCh = new Array(0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0);

	for(var ri = 0; ri < 6; ri++) {
		if(this.row[ri].channel > 0) {
			if(! arCh[this.row[ri].channel-1]++) {
				OutputMidiMsg( -1, 0xE0, this.row[ri].channel,
					dat1, dat2 );
			}
		}
	}

	//PrintLn("PitchWheel out: " + wheelTracker);

	if(wheelTracker == wheelTarget) {
		// At our destination: forget further calls
		if(wheelThread != -1) {
			clearInterval(wheelThread);
			wheelThread = -1;
		}
		return;
	}

}


function ezPatch_fn_plusButtonPress()
{
	//alert("1");
}

function ezPatch_fn_plusButtonRelease()
{
	//alert("2");
}

function ezPatch_fn_minusButtonPress()
{
	//alert("3");
}

function ezPatch_fn_minusButtonRelease()
{
	//alert("4");
}

function ezPatch_fn_soundButtonPress()
{
	//alert("5");
}

function ezPatch_fn_soundButtonRelease()
{
	//alert("6");
}

function ezPatch_fn_songButtonPress()
{
	//alert("7");

	this.panic();
	allNotesOff();
}

function ezPatch_fn_songButtonRelease()
{
	//alert("8");
}



// ===== END ezPatch object =====================================================

